@@ -131,7 +131,7 @@ int Gcalc_function::count_internal(const char *cur_func, uint set_type,
   uint n_ops= c_op & ~(op_any | op_not | v_mask);
   uint n_shape= c_op & ~(op_any | op_not | v_mask); /* same as n_ops */
   op_type v_state= (op_type) (c_op & v_mask);
-  int result= 0, t_counter;
+  int result= 0, t_counter= 0;
   const char *sav_cur_func= cur_func;
 
   // GCALC_DBUG_ENTER("Gcalc_function::count_internal");
@@ -935,6 +935,7 @@ int Gcalc_operation_reducer::count_slice(Gcalc_scan_iterator *si)
         {
           add_poly_border(1, cur_t, prev_state, events);
           prev_state^= 1;
+          prev_range= prev_state ? cur_t : 0;
         }
         if (!events->is_bottom())
         {
