@@ -16,10 +16,6 @@
 
 /* Some general useful functions */
 
-#ifdef USE_PRAGMA_IMPLEMENTATION
-#pragma implementation
-#endif
-
 #include "mariadb.h"
 #include <my_global.h>
 #include <tztime.h>
@@ -34,10 +30,11 @@
 #include "lock.h"
 #include "table.h"
 #include "sql_class.h"
-#include "vers_string.h"
 
 #ifdef WITH_PARTITION_STORAGE_ENGINE
 #include "ha_partition.h"
+#include "sql_table.h"
+#include "transaction.h"
 
 
 partition_info *partition_info::get_clone(THD *thd, bool empty_data_and_index_file)
@@ -178,7 +175,7 @@ bool partition_info::add_named_partition(const char *part_name, size_t length)
   }
   DBUG_PRINT("info", ("Found partition %u is_subpart %d for name %.*s",
                       part_def->part_id, part_def->is_subpart,
-                      length, part_name));
+                      static_cast<int>(length), part_name));
   DBUG_RETURN(false);
 }
 
@@ -304,13 +301,11 @@ bool partition_info::set_partition_bitmaps(List<String> *partition_names)
     The external routine needing this code is check_partition_info
 */
 
-#define MAX_PART_NAME_SIZE 8
-
 char *partition_info::create_default_partition_names(THD *thd, uint part_no,
                                                      uint num_parts_arg,
                                                      uint start_no)
 {
-  char *ptr= (char*) thd->calloc(num_parts_arg * MAX_PART_NAME_SIZE);
+  char *ptr= thd->calloc(num_parts_arg * MAX_PART_NAME_SIZE + 1);
   char *move_ptr= ptr;
   uint i= 0;
   DBUG_ENTER("create_default_partition_names");
@@ -319,7 +314,7 @@ char *partition_info::create_default_partition_names(THD *thd, uint part_no,
   {
     do
     {
-      if (make_partition_name(move_ptr, (start_no + i)))
+      if (!make_partition_name(move_ptr, (start_no + i)).str)
         DBUG_RETURN(NULL);
       move_ptr+= MAX_PART_NAME_SIZE;
     } while (++i < num_parts_arg);
@@ -344,7 +339,7 @@ char *partition_info::create_default_subpartition_name(THD *thd, uint subpart_no
                                                const char *part_name)
 {
   size_t size_alloc= strlen(part_name) + MAX_PART_NAME_SIZE;
-  char *ptr= (char*) thd->calloc(size_alloc);
+  char *ptr= thd->calloc(size_alloc);
   DBUG_ENTER("create_default_subpartition_name");
 
   if (likely(ptr != NULL))
@@ -398,8 +393,6 @@ bool partition_info::set_up_default_partitions(THD *thd, handler *file,
     const char *error_string;
     if (part_type == RANGE_PARTITION)
       error_string= "RANGE";
-    else if (part_type == VERSIONING_PARTITION)
-      error_string= "SYSTEM_TIME";
     else
       error_string= "LIST";
     my_error(ER_PARTITIONS_MUST_BE_DEFINED_ERROR, MYF(0), error_string);
@@ -430,7 +423,7 @@ bool partition_info::set_up_default_partitions(THD *thd, handler *file,
                (!partitions.push_back(part_elem))))
     {
       part_elem->engine_type= default_engine_type;
-      part_elem->partition_name= default_name;
+      part_elem->partition_name= Lex_cstring_strlen(default_name);
       part_elem->id= i;
       default_name+=MAX_PART_NAME_SIZE;
       if (part_type == VERSIONING_PARTITION)
@@ -439,7 +432,7 @@ bool partition_info::set_up_default_partitions(THD *thd, handler *file,
           part_elem->type= partition_element::HISTORY;
         } else {
           part_elem->type= partition_element::CURRENT;
-          part_elem->partition_name= "pn";
+          part_elem->partition_name= Lex_ident_partition("pn"_LEX_CSTRING);
         }
       }
     }
@@ -503,11 +496,11 @@ bool partition_info::set_up_default_subpartitions(THD *thd, handler *file,
           (!part_elem->subpartitions.push_back(subpart_elem))))
       {
         char *ptr= create_default_subpartition_name(thd, j,
-                                                    part_elem->partition_name);
+                                                part_elem->partition_name.str);
         if (!ptr)
           goto end;
         subpart_elem->engine_type= default_engine_type;
-        subpart_elem->partition_name= ptr;
+        subpart_elem->partition_name= Lex_cstring_strlen(ptr);
       }
       else
         goto end;
@@ -565,7 +558,7 @@ bool partition_info::set_up_defaults_for_partitioning(THD *thd, handler *file,
     no parameters
 
   RETURN VALUE
-    Erroneus field name  Error, there are two fields with same name
+    Erroneous field name  Error, there are two fields with same name
     NULL                 Ok, no field defined twice
 
   DESCRIPTION
@@ -575,7 +568,6 @@ bool partition_info::set_up_defaults_for_partitioning(THD *thd, handler *file,
 
 const char* partition_info::find_duplicate_field()
 {
-  const char *field_name_outer, *field_name_inner;
   List_iterator<const char> it_outer(part_field_list);
   uint num_fields= part_field_list.elements;
   uint i,j;
@@ -583,18 +575,16 @@ const char* partition_info::find_duplicate_field()
 
   for (i= 0; i < num_fields; i++)
   {
-    field_name_outer= it_outer++;
+    const Lex_ident_partition field_name_outer= Lex_cstring_strlen(it_outer++);
     List_iterator<const char> it_inner(part_field_list);
     for (j= 0; j < num_fields; j++)
     {
-      field_name_inner= it_inner++;
+      const char *field_name_inner= it_inner++;
       if (i >= j)
         continue;
-      if (!(my_strcasecmp(system_charset_info,
-                          field_name_outer,
-                          field_name_inner)))
+      if (field_name_outer.streq(Lex_cstring_strlen(field_name_inner)))
       {
-        DBUG_RETURN(field_name_outer);
+        DBUG_RETURN(field_name_outer.str);
       }
     }
   }
@@ -617,10 +607,10 @@ const char* partition_info::find_duplicate_field()
   a partition is given for a subpartitioned table, part_elem will be
   the partition, but part_id will be NOT_A_PARTITION_ID and file_name not set.
 */
-partition_element *partition_info::get_part_elem(const char *partition_name,
-                                                 char *file_name,
-                                                 size_t file_name_size,
-                                                 uint32 *part_id)
+partition_element *
+partition_info::get_part_elem(const Lex_ident_partition &partition_name,
+                              char *file_name, size_t file_name_size,
+                              uint32 *part_id)
 {
   List_iterator<partition_element> part_it(partitions);
   uint i= 0;
@@ -637,8 +627,7 @@ partition_element *partition_info::get_part_elem(const char *partition_name,
       do
       {
         partition_element *sub_part_elem= sub_part_it++;
-        if (!my_strcasecmp(system_charset_info,
-                           sub_part_elem->partition_name, partition_name))
+        if (sub_part_elem->partition_name.streq(partition_name))
         {
           if (file_name)
             if (create_subpartition_name(file_name, file_name_size, "",
@@ -651,12 +640,10 @@ partition_element *partition_info::get_part_elem(const char *partition_name,
       } while (++j < num_subparts);
 
       /* Naming a partition (first level) on a subpartitioned table. */
-      if (!my_strcasecmp(system_charset_info,
-                            part_elem->partition_name, partition_name))
+      if (part_elem->partition_name.streq(partition_name))
         DBUG_RETURN(part_elem);
     }
-    else if (!my_strcasecmp(system_charset_info,
-                            part_elem->partition_name, partition_name))
+    else if (part_elem->partition_name.streq(partition_name))
     {
       if (file_name)
         if (create_partition_name(file_name, file_name_size, "",
@@ -674,11 +661,11 @@ partition_element *partition_info::get_part_elem(const char *partition_name,
   Helper function to find_duplicate_name.
 */
 
-static const char *get_part_name_from_elem(const char *name, size_t *length,
-                                      my_bool not_used __attribute__((unused)))
+static const uchar *get_part_name_from_elem(const void *name, size_t *length,
+                                            my_bool)
 {
-  *length= strlen(name);
-  return name;
+  *length= strlen(static_cast<const char *>(name));
+  return static_cast<const uchar *>(name);
 }
 
 /*
@@ -716,8 +703,9 @@ char *partition_info::find_duplicate_name()
   max_names= num_parts;
   if (is_sub_partitioned())
     max_names+= num_parts * num_subparts;
-  if (my_hash_init(PSI_INSTRUMENT_ME, &partition_names, system_charset_info, max_names, 0, 0,
-                   (my_hash_get_key) get_part_name_from_elem, 0, HASH_UNIQUE))
+  if (my_hash_init(PSI_INSTRUMENT_ME, &partition_names,
+                   Lex_ident_partition::charset_info(),
+                   max_names, 0, 0, get_part_name_from_elem, 0, HASH_UNIQUE))
   {
     DBUG_ASSERT(0);
     curr_name= (const uchar*) "Internal failure";
@@ -725,7 +713,7 @@ char *partition_info::find_duplicate_name()
   }
   while ((p_elem= (parts_it++)))
   {
-    curr_name= (const uchar*) p_elem->partition_name;
+    curr_name= (const uchar*) p_elem->partition_name.str;
     if (my_hash_insert(&partition_names, curr_name))
       goto error;
 
@@ -735,7 +723,7 @@ char *partition_info::find_duplicate_name()
       partition_element *subp_elem;
       while ((subp_elem= (subparts_it++)))
       {
-        curr_name= (const uchar*) subp_elem->partition_name;
+        curr_name= (const uchar*) subp_elem->partition_name.str;
         if (my_hash_insert(&partition_names, curr_name))
           goto error;
       }
@@ -765,15 +753,14 @@ bool partition_info::has_unique_name(partition_element *element)
 {
   DBUG_ENTER("partition_info::has_unique_name");
   
-  const char *name_to_check= element->partition_name;
   List_iterator<partition_element> parts_it(partitions);
   
   partition_element *el;
   while ((el= (parts_it++)))
   {
-    if (!(my_strcasecmp(system_charset_info, el->partition_name, 
-                        name_to_check)) && el != element)
-        DBUG_RETURN(FALSE);
+    if (element->partition_name.streq(el->partition_name) &&
+        el != element)
+      DBUG_RETURN(FALSE);
 
     if (!el->subpartitions.is_empty()) 
     {
@@ -781,9 +768,9 @@ bool partition_info::has_unique_name(partition_element *element)
       List_iterator<partition_element> subparts_it(el->subpartitions);
       while ((sub_el= (subparts_it++)))
       {
-        if (!(my_strcasecmp(system_charset_info, sub_el->partition_name, 
-                            name_to_check)) && sub_el != element)
-            DBUG_RETURN(FALSE);
+        if (element->partition_name.streq(sub_el->partition_name) &&
+            sub_el != element)
+          DBUG_RETURN(FALSE);
       }
     }
   } 
@@ -799,18 +786,13 @@ bool partition_info::has_unique_name(partition_element *element)
     vers_info->interval   Limit by fixed time interval
     vers_info->hist_part  (out) Working history partition
 */
-int partition_info::vers_set_hist_part(THD *thd)
+bool partition_info::vers_set_hist_part(THD *thd, uint *create_count)
 {
-  if (!vers_require_hist_part(thd))
-    return 0;
+  const bool auto_hist= create_count && vers_info->auto_hist;
 
-  if (table->pos_in_table_list &&
-      table->pos_in_table_list->partition_names)
-  {
-    return HA_ERR_PARTITION_LIST;
-  }
   if (vers_info->limit)
   {
+    DBUG_ASSERT(!vers_info->interval.is_set());
     ha_partition *hp= (ha_partition*)(table->file);
     partition_element *next;
     List_iterator<partition_element> it(partitions);
@@ -825,17 +807,25 @@ int partition_info::vers_set_hist_part(THD *thd)
       vers_info->hist_part= next;
       records= next_records;
     }
-    if (records >= vers_info->limit && next != vers_info->now_part)
-      vers_info->hist_part= next;
+    if (records >= vers_info->limit)
+    {
+      if (next == vers_info->now_part)
+      {
+        if (auto_hist)
+          *create_count= 1;
+      }
+      else
+        vers_info->hist_part= next;
+    }
     return 0;
   }
-
-  if (vers_info->interval.is_set())
+  else if (vers_info->interval.is_set() &&
+           /* Left boundary is closed */
+           vers_info->hist_part->range_value <= (longlong)thd->query_start() &&
+           vers_info->hist_part->range_value < TIMESTAMP_MAX_VALUE)
   {
-    if (vers_info->hist_part->range_value > thd->query_start())
-      return 0;
-
     partition_element *next= NULL;
+    bool error= true;
     List_iterator<partition_element> it(partitions);
     while (next != vers_info->hist_part)
       next= it++;
@@ -843,11 +833,174 @@ int partition_info::vers_set_hist_part(THD *thd)
     while ((next= it++) != vers_info->now_part)
     {
       vers_info->hist_part= next;
-      if (next->range_value > thd->query_start())
-        return 0;
+      /* Right boundary is open */
+      if ((longlong)thd->query_start() < next->range_value)
+      {
+        error= false;
+        break;
+      }
+    }
+    if (error)
+    {
+      if (auto_hist)
+      {
+        *create_count= 0;
+        const my_time_t hist_end= (my_time_t) vers_info->hist_part->range_value;
+        DBUG_ASSERT(thd->query_start() >= hist_end);
+        MYSQL_TIME h0, q0;
+        my_tz_OFFSET0->gmt_sec_to_TIME(&h0, hist_end);
+        my_tz_OFFSET0->gmt_sec_to_TIME(&q0, thd->query_start());
+        longlong q= pack_time(&q0);
+        longlong h= pack_time(&h0);
+        while (h <= q)
+        {
+          if (date_add_interval(thd, &h0, vers_info->interval.type,
+                                vers_info->interval.step))
+            return true;
+          h= pack_time(&h0);
+          ++*create_count;
+          if (*create_count == MAX_PARTITIONS - 2)
+          {
+            my_error(ER_TOO_MANY_PARTITIONS_ERROR, MYF(ME_WARNING));
+            my_error(ER_VERS_HIST_PART_FAILED, MYF(0),
+                     table->s->db.str, table->s->table_name.str);
+            return true;
+          }
+        }
+      }
+      else
+      {
+        my_error(WARN_VERS_PART_FULL, MYF(ME_WARNING|ME_ERROR_LOG),
+                table->s->db.str, table->s->table_name.str,
+                vers_info->hist_part->partition_name.str, "INTERVAL");
+      }
     }
   }
-  return 0;
+
+  return false;
+}
+
+
+/**
+  @brief Run fast_alter_partition_table() to add new history partitions
+         for tables requiring them.
+
+  @param num_parts  Number of partitions to create
+*/
+bool vers_create_partitions(THD *thd, TABLE_LIST* tl, uint num_parts)
+{
+  bool result= true;
+  Table_specification_st create_info;
+  Alter_info alter_info;
+  partition_info *save_part_info= thd->work_part_info;
+  Query_tables_list save_query_tables;
+  Reprepare_observer *save_reprepare_observer= thd->m_reprepare_observer;
+  bool save_no_write_to_binlog= thd->lex->no_write_to_binlog;
+  thd->m_reprepare_observer= NULL;
+  thd->lex->reset_n_backup_query_tables_list(&save_query_tables);
+  thd->lex->no_write_to_binlog= true;
+  TABLE *table= tl->table;
+
+  DBUG_ASSERT(!thd->is_error());
+  DBUG_ASSERT(num_parts);
+
+  {
+    DBUG_ASSERT(table->s->get_table_ref_type() == TABLE_REF_BASE_TABLE);
+    DBUG_ASSERT(table->versioned());
+    DBUG_ASSERT(table->part_info);
+    DBUG_ASSERT(table->part_info->vers_info);
+    alter_info.reset();
+    alter_info.partition_flags= ALTER_PARTITION_ADD|ALTER_PARTITION_AUTO_HIST;
+    create_info.init();
+    create_info.alter_info= &alter_info;
+    Alter_table_ctx alter_ctx(thd, tl, 1, &table->s->db, &table->s->table_name);
+
+    MDL_REQUEST_INIT(&tl->mdl_request, MDL_key::TABLE, tl->db.str,
+                    tl->table_name.str, MDL_SHARED_NO_WRITE, MDL_TRANSACTION);
+    if (thd->mdl_context.acquire_lock(&tl->mdl_request,
+                                      thd->variables.lock_wait_timeout))
+      goto exit;
+    table->mdl_ticket= tl->mdl_request.ticket;
+
+    create_info.db_type= table->s->db_type();
+    create_info.options|= HA_VERSIONED_TABLE;
+    DBUG_ASSERT(create_info.db_type);
+
+    create_info.vers_info.set_start(table->s->vers_start_field()->field_name);
+    create_info.vers_info.set_end(table->s->vers_end_field()->field_name);
+
+    partition_info *part_info= new partition_info();
+    if (unlikely(!part_info))
+    {
+      my_error(ER_OUT_OF_RESOURCES, MYF(0));
+      goto exit;
+    }
+    part_info->use_default_num_partitions= false;
+    part_info->use_default_num_subpartitions= false;
+    part_info->num_parts= num_parts;
+    part_info->num_subparts= table->part_info->num_subparts;
+    part_info->subpart_type= table->part_info->subpart_type;
+    if (unlikely(part_info->vers_init_info(thd)))
+    {
+      my_error(ER_OUT_OF_RESOURCES, MYF(0));
+      goto exit;
+    }
+
+    thd->work_part_info= part_info;
+    if (part_info->set_up_defaults_for_partitioning(thd, table->file, NULL,
+                                    table->part_info->next_part_no(num_parts)))
+    {
+      my_error(ER_VERS_HIST_PART_FAILED, MYF(ME_WARNING),
+               tl->db.str, tl->table_name.str);
+      goto exit;
+    }
+    bool partition_changed= false;
+    bool fast_alter_partition= false;
+    if (prep_alter_part_table(thd, table, &alter_info, &create_info,
+                              &partition_changed, &fast_alter_partition))
+    {
+      my_error(ER_VERS_HIST_PART_FAILED, MYF(ME_WARNING),
+               tl->db.str, tl->table_name.str);
+      goto exit;
+    }
+    if (!fast_alter_partition)
+    {
+      my_error(ER_VERS_HIST_PART_FAILED, MYF(ME_WARNING),
+               tl->db.str, tl->table_name.str);
+      goto exit;
+    }
+    DBUG_ASSERT(partition_changed);
+    if (mysql_prepare_alter_table(thd, table, &create_info, &alter_info,
+                                  &alter_ctx))
+    {
+      my_error(ER_VERS_HIST_PART_FAILED, MYF(ME_WARNING),
+               tl->db.str, tl->table_name.str);
+      goto exit;
+    }
+
+    alter_info.db= alter_ctx.db;
+    alter_info.table_name= alter_ctx.table_name;
+    if (fast_alter_partition_table(thd, table, &alter_info, &alter_ctx,
+                                   &create_info, tl))
+    {
+      my_error(ER_VERS_HIST_PART_FAILED, MYF(ME_WARNING),
+               tl->db.str, tl->table_name.str);
+      goto exit;
+    }
+  }
+
+  result= false;
+  // NOTE: we have to return DA_EMPTY for new command
+  DBUG_ASSERT(thd->get_stmt_da()->is_ok());
+  thd->get_stmt_da()->reset_diagnostics_area();
+  thd->variables.option_bits|= OPTION_BINLOG_THIS;
+
+exit:
+  thd->work_part_info= save_part_info;
+  thd->m_reprepare_observer= save_reprepare_observer;
+  thd->lex->restore_backup_query_tables_list(&save_query_tables);
+  thd->lex->no_write_to_binlog= save_no_write_to_binlog;
+  return result;
 }
 
 
@@ -856,7 +1009,7 @@ int partition_info::vers_set_hist_part(THD *thd)
 */
 void partition_info::vers_check_limit(THD *thd)
 {
-  if (!vers_info->limit ||
+  if (vers_info->auto_hist || !vers_info->limit ||
       vers_info->hist_part->id + 1 < vers_info->now_part->id)
     return;
 
@@ -881,11 +1034,11 @@ void partition_info::vers_check_limit(THD *thd)
                         WARN_VERS_PART_FULL,
                         ER_THD(thd, WARN_VERS_PART_FULL),
                         table->s->db.str, table->s->table_name.str,
-                        vers_info->hist_part->partition_name, "LIMIT");
+                        vers_info->hist_part->partition_name.str, "LIMIT");
 
-    sql_print_warning(ER_THD(thd, WARN_VERS_PART_FULL),
+    sql_print_warning(ER_DEFAULT(WARN_VERS_PART_FULL),
                       table->s->db.str, table->s->table_name.str,
-                      vers_info->hist_part->partition_name, "LIMIT");
+                      vers_info->hist_part->partition_name.str, "LIMIT");
   }
 }
 
@@ -1122,15 +1275,14 @@ bool partition_info::check_partition_info(THD *thd, handlerton **eng_type,
     if (!list_of_part_fields)
     {
       DBUG_ASSERT(part_expr);
-      err= part_expr->walk(&Item::check_partition_func_processor, 0, NULL);
+      err= part_expr->walk(&Item::check_partition_func_processor, 0, 0);
     }
 
     /* Check for sub partition expression. */
     if (!err && is_sub_partitioned() && !list_of_subpart_fields)
     {
       DBUG_ASSERT(subpart_expr);
-      err= subpart_expr->walk(&Item::check_partition_func_processor, 0,
-                              NULL);
+      err= subpart_expr->walk(&Item::check_partition_func_processor, 0, 0);
     }
 
     if (err)
@@ -1242,8 +1394,7 @@ bool partition_info::check_partition_info(THD *thd, handlerton **eng_type,
           num_parts_not_set++;
           part_elem->engine_type= default_engine_type;
         }
-        if (check_table_name(part_elem->partition_name,
-                             strlen(part_elem->partition_name), FALSE))
+        if (Lex_ident_table::check_name(part_elem->partition_name, false))
         {
           my_error(ER_WRONG_PARTITION_NAME, MYF(0));
           goto end;
@@ -1261,8 +1412,7 @@ bool partition_info::check_partition_info(THD *thd, handlerton **eng_type,
         {
           sub_elem= sub_it++;
           warn_if_dir_in_part_elem(thd, sub_elem);
-          if (check_table_name(sub_elem->partition_name,
-                               strlen(sub_elem->partition_name), FALSE))
+          if (Lex_ident_table::check_name(sub_elem->partition_name, false))
           {
             my_error(ER_WRONG_PARTITION_NAME, MYF(0));
             goto end;
@@ -1504,7 +1654,6 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
   Field *field, **ptr;
   uchar **char_ptrs;
   unsigned i;
-  size_t size;
   uint tot_part_fields= 0;
   uint tot_subpart_fields= 0;
   DBUG_ENTER("set_up_charset_field_preps");
@@ -1518,15 +1667,13 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
     while ((field= *(ptr++)))
       if (field_is_partition_charset(field))
         tot_part_fields++;
-    size= tot_part_fields * sizeof(char*);
-    if (!(char_ptrs= (uchar**)thd->calloc(size)))
+    if (!(char_ptrs= thd->calloc<uchar*>(tot_part_fields)))
       goto error;
     part_field_buffers= char_ptrs;
-    if (!(char_ptrs= (uchar**)thd->calloc(size)))
+    if (!(char_ptrs= thd->calloc<uchar*>(tot_part_fields)))
       goto error;
     restore_part_field_ptrs= char_ptrs;
-    size= (tot_part_fields + 1) * sizeof(Field*);
-    if (!(char_ptrs= (uchar**)thd->alloc(size)))
+    if (!(char_ptrs= thd->alloc<uchar*>(tot_part_fields + 1)))
       goto error;
     part_charset_field_array= (Field**)char_ptrs;
     ptr= part_field_array;
@@ -1536,8 +1683,7 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
       if (field_is_partition_charset(field))
       {
         uchar *field_buf;
-        size= field->pack_length();
-        if (!(field_buf= (uchar*) thd->calloc(size)))
+        if (!(field_buf= thd->calloc<uchar>(field->pack_length())))
           goto error;
         part_charset_field_array[i]= field;
         part_field_buffers[i++]= field_buf;
@@ -1553,15 +1699,13 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
     while ((field= *(ptr++)))
       if (field_is_partition_charset(field))
         tot_subpart_fields++;
-    size= tot_subpart_fields * sizeof(char*);
-    if (!(char_ptrs= (uchar**) thd->calloc(size)))
+    if (!(char_ptrs= (uchar**) thd->calloc<uchar*>(tot_subpart_fields)))
       goto error;
     subpart_field_buffers= char_ptrs;
-    if (!(char_ptrs= (uchar**) thd->calloc(size)))
+    if (!(char_ptrs= (uchar**) thd->calloc<uchar*>(tot_subpart_fields)))
       goto error;
     restore_subpart_field_ptrs= char_ptrs;
-    size= (tot_subpart_fields + 1) * sizeof(Field*);
-    if (!(char_ptrs= (uchar**) thd->alloc(size)))
+    if (!(char_ptrs= (uchar**) thd->alloc<uchar*>(tot_subpart_fields + 1)))
       goto error;
     subpart_charset_field_array= (Field**)char_ptrs;
     ptr= subpart_field_array;
@@ -1572,8 +1716,7 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
 
       if (!field_is_partition_charset(field))
         continue;
-      size= field->pack_length();
-      if (!(field_buf= (uchar*) thd->calloc(size)))
+      if (!(field_buf= thd->calloc<uchar>(field->pack_length())))
         goto error;
       subpart_charset_field_array[i]= field;
       subpart_field_buffers[i++]= field_buf;
@@ -1824,7 +1967,7 @@ bool partition_info::add_column_list_value(THD *thd, Item *item)
   part_column_list_val *col_val;
   Name_resolution_context *context= &thd->lex->current_select->context;
   TABLE_LIST *save_list= context->table_list;
-  const char *save_where= thd->where;
+  THD_WHERE save_where= thd->where;
   DBUG_ENTER("partition_info::add_column_list_value");
 
   if (part_type == LIST_PARTITION &&
@@ -1838,11 +1981,11 @@ bool partition_info::add_column_list_value(THD *thd, Item *item)
 
   context->table_list= 0;
   if (column_list)
-    thd->where= "field list";
+    thd->where= THD_WHERE::FIELD_LIST;
   else
-    thd->where= "partition function";
+    thd->where= THD_WHERE::PARTITION_FUNCTION;
 
-  if (item->walk(&Item::check_partition_func_processor, 0, NULL))
+  if (item->walk(&Item::check_partition_func_processor, 0, 0))
   {
     my_error(ER_PARTITION_FUNCTION_IS_NOT_ALLOWED, MYF(0));
     DBUG_RETURN(TRUE);
@@ -1887,8 +2030,7 @@ bool partition_info::init_column_part(THD *thd)
   uint loc_num_columns;
   DBUG_ENTER("partition_info::init_column_part");
 
-  if (!(list_val=
-      (part_elem_value*) thd->calloc(sizeof(part_elem_value))) ||
+  if (!(list_val= thd->calloc<part_elem_value>(1)) ||
       p_elem->list_val_list.push_back(list_val, thd->mem_root))
     DBUG_RETURN(TRUE);
 
@@ -1896,9 +2038,7 @@ bool partition_info::init_column_part(THD *thd)
     loc_num_columns= num_columns;
   else
     loc_num_columns= MAX_REF_PARTS;
-  if (!(col_val_array=
-        (part_column_list_val*) thd->calloc(loc_num_columns *
-                                            sizeof(part_column_list_val))))
+  if (!(col_val_array= thd->calloc<part_column_list_val>(loc_num_columns)))
     DBUG_RETURN(TRUE);
 
   list_val->col_val_array= col_val_array;
@@ -2003,7 +2143,7 @@ int partition_info::fix_partition_values(THD *thd,
   else if (item_expr->result_type() != INT_RESULT)
   {
     my_error(ER_VALUES_IS_NOT_INT_TYPE_ERROR, MYF(0),
-             part_elem->partition_name);
+             part_elem->partition_name.str);
     DBUG_RETURN(TRUE);
   }
   if (part_type == RANGE_PARTITION)
@@ -2397,9 +2537,8 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
   while ((old_name= old_field_name_it++))
   {
     new_name= new_field_name_it++;
-    if (!new_name || my_strcasecmp(system_charset_info,
-                                   new_name,
-                                   old_name))
+    if (!new_name || !Lex_ident_partition(Lex_cstring_strlen(new_name)).
+                        streq(Lex_cstring_strlen(old_name)))
       DBUG_RETURN(false);
   }
 
@@ -2412,9 +2551,8 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
     while ((old_name= old_field_name_it++))
     {
       new_name= new_field_name_it++;
-      if (!new_name || my_strcasecmp(system_charset_info,
-                                     new_name,
-                                     old_name))
+      if (!new_name || !Lex_ident_partition(Lex_cstring_strlen(new_name)).
+                          streq(Lex_cstring_strlen(old_name)))
         DBUG_RETURN(false);
     }
   }
@@ -2443,8 +2581,8 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
         part_state must be PART_NORMAL!
       */
       if (!part_elem || !new_part_elem ||
-          strcmp(part_elem->partition_name,
-                 new_part_elem->partition_name) ||
+          strcmp(part_elem->partition_name.str,
+                 new_part_elem->partition_name.str) ||
           part_elem->part_state != PART_NORMAL ||
           new_part_elem->part_state != PART_NORMAL ||
           part_elem->max_value != new_part_elem->max_value ||
@@ -2505,8 +2643,8 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
                 sub_part_elem->engine_type != new_sub_part_elem->engine_type)
               DBUG_RETURN(false);
 
-            if (strcmp(sub_part_elem->partition_name,
-                       new_sub_part_elem->partition_name) ||
+            if (strcmp(sub_part_elem->partition_name.str,
+                       new_sub_part_elem->partition_name.str) ||
                 sub_part_elem->part_state != PART_NORMAL ||
                 new_sub_part_elem->part_state != PART_NORMAL ||
                 sub_part_elem->part_min_rows !=
@@ -2650,13 +2788,14 @@ bool partition_info::vers_init_info(THD * thd)
 
 bool partition_info::vers_set_interval(THD* thd, Item* interval,
                                        interval_type int_type, Item* starts,
-                                       const char *table_name)
+                                       bool auto_hist, const char *table_name)
 {
   DBUG_ASSERT(part_type == VERSIONING_PARTITION);
 
   MYSQL_TIME ltime;
   uint err;
   vers_info->interval.type= int_type;
+  vers_info->auto_hist= auto_hist;
 
   /* 1. assign INTERVAL to interval.step */
   if (interval->fix_fields_if_needed_for_scalar(thd, &interval))
@@ -2683,7 +2822,7 @@ bool partition_info::vers_set_interval(THD* thd, Item* interval,
       case DECIMAL_RESULT:
       case REAL_RESULT:
         /* When table member is defined, we are inside mysql_unpack_partition(). */
-        if (!table || starts->val_int() > TIMESTAMP_MAX_VALUE)
+        if (!table || (ulonglong) starts->val_int() > TIMESTAMP_MAX_VALUE)
           goto interval_starts_error;
         vers_info->interval.start= (my_time_t) starts->val_int();
         break;
@@ -2738,6 +2877,23 @@ interval_starts_error:
 }
 
 
+bool partition_info::vers_set_limit(ulonglong limit, bool auto_hist,
+                                    const char *table_name)
+{
+  DBUG_ASSERT(part_type == VERSIONING_PARTITION);
+
+  if (limit < 1)
+  {
+    my_error(ER_PART_WRONG_VALUE, MYF(0), table_name, "LIMIT");
+    return true;
+  }
+
+  vers_info->limit= limit;
+  vers_info->auto_hist= auto_hist;
+  return !limit;
+}
+
+
 bool partition_info::error_if_requires_values() const
 {
   switch (part_type) {
